import datetime

from cpl_core.configuration import ConfigurationABC

from bot_data.abc.achievement_repository_abc import AchievementRepositoryABC
from bot_data.abc.user_joined_game_server_repository_abc import UserJoinedGameServerRepositoryABC
from bot_data.abc.user_joined_voice_channel_repository_abc import UserJoinedVoiceChannelRepositoryABC
from bot_data.abc.user_message_count_per_hour_repository_abc import UserMessageCountPerHourRepositoryABC
from bot_data.abc.user_repository_abc import UserRepositoryABC
from bot_data.abc.user_warnings_repository_abc import UserWarningsRepositoryABC
from bot_data.model.server_config import ServerConfig
from bot_graphql.abc.query_abc import QueryABC


class ServerStatisticQuery(QueryABC):
    def __init__(
        self,
        config: ConfigurationABC,
        users: UserRepositoryABC,
        user_joined_voice_channels: UserJoinedVoiceChannelRepositoryABC,
        user_joined_game_servers: UserJoinedGameServerRepositoryABC,
        user_messages: UserMessageCountPerHourRepositoryABC,
        user_warnings: UserWarningsRepositoryABC,
        achievements: AchievementRepositoryABC,
    ):
        QueryABC.__init__(self, "ServerStatistic")

        self._config = config
        self._users = users
        self._user_joined_voice_channels = user_joined_voice_channels
        self._user_joined_game_servers = user_joined_game_servers
        self._user_messages = user_messages
        self._user_warnings = user_warnings
        self._achievements = achievements

        self.set_field(
            "achievementsAchieved",
            self._resolve_achievements,
        )

        self.set_field(
            "messageCount",
            self._resolve_message_count,
        )

        self.set_field("userCount", lambda server, *_: self._users.get_users_by_server_id(server.server.id).count())
        self.set_field("activeUserCount", self._resolve_active_count)

        self.set_field("userJoinedVoiceChannelCount", self._resolve_voice_channel_count)
        self.set_field("userJoinedVoiceChannelOntime", self._resolve_voice_channel_ontime)

        self.set_field("userJoinedGameServerCount", self._resolve_game_server_count)
        self.set_field("userJoinedGameServerOntime", self._resolve_game_server_ontime)

        self.set_field("userWarningCount", self._resolve_user_warning_count)

        self.set_field("activityScore", self._resolve_activity_score)

    def _resolve_activity_score(self, server, *_):
        days = (datetime.date.today() - self._get_date(**server.kwargs)).days
        return int(
            (
                (
                    self._resolve_achievements(server, *_)
                    + self._resolve_message_count(server, *_)
                    + self._resolve_voice_channel_count(server, *_)
                    + self._resolve_voice_channel_ontime(server, *_)
                    + self._resolve_game_server_count(server, *_)
                    + self._resolve_game_server_ontime(server, *_)
                    - self._resolve_user_warning_count(server, *_)
                )
                / self._resolve_active_count(server, *_)
            )
            / days
            * 1000
        )

    def _resolve_message_count(self, server, *_):
        settings: ServerConfig = self._config.get_configuration(f"ServerConfig_{server.server.discord_id}")
        return (
            self._user_messages.get_user_message_count_per_hours()
            .where(
                lambda x: x.user.server.id == server.server.id
                and x.created_at.date() >= self._get_date(**server.kwargs)
            )
            .sum(lambda x: x.xp_count / settings.xp_per_message)
        )

    def _resolve_active_count(self, server, *_):
        return self._users.get_users_by_server_id(server.server.id).where(lambda x: not x.left_server).count()

    def _resolve_voice_channel_count(self, server, *_):
        return (
            self._user_joined_voice_channels.get_user_joined_voice_channels()
            .where(
                lambda x: x.user.server.id == server.server.id
                and x.created_at.date() >= self._get_date(**server.kwargs)
            )
            .count()
        )

    def _resolve_voice_channel_ontime(self, server, *_):
        return (
            self._user_joined_voice_channels.get_user_joined_voice_channels()
            .where(
                lambda x: x.user.server.id == server.server.id
                and x.created_at.date() >= self._get_date(**server.kwargs)
            )
            .sum(lambda x: x.time)
        )

    def _resolve_game_server_count(self, server, *_):
        return (
            self._user_joined_game_servers.get_user_joined_game_servers()
            .where(
                lambda x: x.user.server.id == server.server.id
                and x.created_at.date() >= self._get_date(**server.kwargs)
            )
            .count()
        )

    def _resolve_game_server_ontime(self, server, *_):
        return (
            self._user_joined_game_servers.get_user_joined_game_servers()
            .where(
                lambda x: x.user.server.id == server.server.id
                and x.created_at.date() >= self._get_date(**server.kwargs)
            )
            .sum(lambda x: x.time)
        )

    def _resolve_user_warning_count(self, server, *_):
        return (
            self._user_warnings.get_user_warnings()
            .where(
                lambda x: x.user.server.id == server.server.id
                and x.created_at.date() >= self._get_date(**server.kwargs)
            )
            .count()
        )

    def _resolve_achievements(self, server, *_):
        return (
            self._achievements.get_achievements_by_server_id(server.server.id)
            .select_many(lambda x: self._achievements.get_user_got_achievements_by_achievement_id(x.id))
            .where(
                lambda x: x.user.server.id == server.server.id
                and x.created_at.date() >= self._get_date(**server.kwargs)
            )
            .count()
        )

    def _get_date(self, **kwargs) -> datetime.date:
        if "date" not in kwargs:
            return datetime.date.today() - datetime.timedelta(days=7)

        return datetime.datetime.strptime(kwargs["date"], "%d.%m.%Y").date()
